/*-
 * Copyright (c) 2006 Hajimu UMEMOTO <ume@mahoroba.org>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__RCSID("$Mahoroba: src/nss_resinit/nss_resinit.c,v 1.9 2006/04/12 11:29:35 ume Exp $");

#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <errno.h>
#include <kenv.h>
#include <nss.h>
#include <pthread.h>
#include <pthread_np.h>
#include <netdb.h>
#include <resolv.h>
#include <stdlib.h>

#ifndef res_ninit
typedef struct __res_state *res_state;
#define	res_ninit(res)			res_init()
#define	__res_vinit(res, preinit)	(res_close(), res_init())
#endif

#define	timespeccmp(tvp, uvp, cmp)					\
	(((tvp)->tv_sec == (uvp)->tv_sec) ?				\
	    ((tvp)->tv_nsec cmp (uvp)->tv_nsec) :			\
	    ((tvp)->tv_sec cmp (uvp)->tv_sec))

#define	CHECK_INTERVAL	10

ns_mtab *nss_module_register(const char *, unsigned int *,
    nss_module_unregister_fn *);

static NSS_METHOD_PROTOTYPE(nss_wrapper_hosts);

static ns_mtab methods[] = {
	{ NSDB_HOSTS, "getaddrinfo",	  nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "ghbyname",	  nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "ghbyaddr",	  nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "gethostbyname2_r", nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "gethostbyaddr_r",  nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "getnetbyname_r",	  nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "getnetbyaddr_r",	  nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "gethostbyname",	  nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "gethostbyaddr",	  nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "getnetbyname",	  nss_wrapper_hosts, NULL },
	{ NSDB_HOSTS, "getnetbyaddr",	  nss_wrapper_hosts, NULL }
};

static time_t check_interval = -1;

struct timestamp {
	struct timespec mtimespec;
	time_t chktime;
};

static struct timestamp timestamp;
static pthread_key_t timestamp_key;
static pthread_once_t timestamp_init_once = PTHREAD_ONCE_INIT;
static int timestamp_key_created = 0;

static void
timestamp_keycreate(void)
{
	timestamp_key_created =
	    (pthread_key_create(&timestamp_key, free) == 0);
}

static struct timestamp *
timestamp_init(void)
{
	struct timestamp *ts;

	if (pthread_main_np() != 0)
		return (&timestamp);
	if (pthread_once(&timestamp_init_once, timestamp_keycreate) != 0 ||
	    !timestamp_key_created)
		return (NULL);
	if ((ts = pthread_getspecific(timestamp_key)) != NULL)
		return (ts);
	if ((ts = calloc(1, sizeof(*ts))) == NULL)
		return (NULL);
	if (pthread_setspecific(timestamp_key, ts) == 0)
		return (ts);
	free(ts);
	return (NULL);
}

static void
setinterval(struct timestamp *ts)
{
	time_t interval;
	char buf[6], *ep;

	if (check_interval >= 0)
		return;
	if (kenv(KENV_GET, "nss_resinit_interval", buf, sizeof(buf)) == -1) {
		check_interval = CHECK_INTERVAL;
		return;
	}
	errno = 0;
	interval = (time_t)strtol(buf, &ep, 10);
	if (errno == 0 && ep && *ep == '\0' && interval >= 0)
		check_interval = interval;
	else
		check_interval = CHECK_INTERVAL;
}

static int
res_initialize(res_state statp)
{
	struct timestamp *ts;
	struct stat sb;
	time_t now, chktime;
	int rval, stat_rval;

	if ((ts = timestamp_init()) == NULL)
		return (-1);
	setinterval(ts);

	time(&now);

	if ((statp->options & RES_INIT) == 0) {
		stat_rval = stat(_PATH_RESCONF, &sb);
		rval = res_ninit(statp);
		if (rval != -1 && stat_rval != -1)
			ts->mtimespec = sb.st_mtimespec;
		ts->chktime = now + check_interval;
		return (rval);
	}

	chktime = ts->chktime;
	if (now < chktime)
		return (0);
	ts->chktime = now + check_interval;

	if (stat(_PATH_RESCONF, &sb) == -1) {
		/*
		 * Lost the file, in chroot?
		 * Don' trash settings
		 */
		return (0);
	}

	/*
	 * XXX: The timestamp is not initialized, yet.  It occurs when
	 * res_init(3) was already called before nss_resinit.
	 * This is a workaround to avoid unnecessarily duplicate call
	 * of res_init(3).
	 * There is race where _res might not be re-initialized even
	 * when resolv.conf was updated.  To fix this issue, we need
	 * to change res_init(3) to set timestamp.
	 */
	if (chktime == 0) {
		ts->mtimespec = sb.st_mtimespec;
		return (0);
	}

	if (timespeccmp(&sb.st_mtimespec, &ts->mtimespec, !=)) {
		rval = __res_vinit(statp, 1);
		if (rval != -1)
			ts->mtimespec = sb.st_mtimespec;
		return (rval);
	}
	return (0);
}

static int
nss_wrapper_hosts(void *retval __unused, void *mdata __unused,
    va_list ap __unused)
{
	if (res_initialize(&_res) == -1)
		return (NS_UNAVAIL);
	return (NS_NOTFOUND);
}

ns_mtab *
nss_module_register(const char *source __unused, unsigned int *mtabsize,
    nss_module_unregister_fn *unreg)
{
	*mtabsize = sizeof(methods) / sizeof(methods[0]);
	*unreg = NULL;
	return (methods);
}
